import json
import logging
import os
import sys
from typing import List
from urllib.parse import urlparse

import click
import yaml
from meltano.core.plugin import PluginType
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.project_add_service import ProjectAddService
from meltano.core.project_plugins_service import ProjectPluginsService

from . import cli
from .params import pass_project
from .utils import CliError, add_plugin, add_related_plugins, install_plugins


@cli.command()
@click.argument("plugin_type", type=click.Choice(PluginType.cli_arguments()))
@click.argument("plugin_name", nargs=-1, required=True)
@click.option("--inherit-from")
@click.option("--variant")
@click.option("--as", "as_name")
@click.option("--custom", is_flag=True)
@click.option("--include-related", is_flag=True)
@pass_project()
@click.pass_context
def add(
    ctx,
    project,
    plugin_type,
    plugin_name,
    inherit_from=None,
    variant=None,
    as_name=None,
    **flags,
):
    """Add a plugin to your project."""
    plugin_type = PluginType.from_cli_argument(plugin_type)
    plugin_names = plugin_name  # nargs=-1

    if as_name:
        # `add <type> <inherit-from> --as <name>``
        # is equivalent to:
        # `add <type> <name> --inherit-from <inherit-from>``
        inherit_from = plugin_names[0]
        plugin_names = [as_name]

    plugins_service = ProjectPluginsService(project)

    if flags["custom"]:
        if plugin_type in (
            PluginType.TRANSFORMERS,
            PluginType.TRANSFORMS,
            PluginType.ORCHESTRATORS,
        ):
            raise CliError(f"--custom is not supported for {plugin_type}")

    add_service = ProjectAddService(project, plugins_service=plugins_service)

    plugins = [
        add_plugin(
            project,
            plugin_type,
            plugin_name,
            inherit_from=inherit_from,
            variant=variant,
            custom=flags["custom"],
            add_service=add_service,
        )
        for plugin_name in plugin_names
    ]

    related_plugin_types = [PluginType.FILES]
    if flags["include_related"]:
        related_plugin_types = list(PluginType)

    related_plugins = add_related_plugins(
        project, plugins, add_service=add_service, plugin_types=related_plugin_types
    )
    plugins.extend(related_plugins)

    # We will install the plugins in reverse order, since dependencies
    # are listed after their dependents in `related_plugins`, but should
    # be installed first.
    plugins.reverse()

    # Plugin installation can be order dependent (e.g. a dbt transform package
    # requires the dbt transformer to be installed before), so we will disable
    # parallelism for this operation.
    success = install_plugins(
        project,
        plugins,
        reason=PluginInstallReason.ADD,
        parallelism=1,
    )

    if not success:
        raise CliError("Failed to install plugin(s)")

    _print_plugins(plugins)


def _print_plugins(plugins):
    printed_empty_line = False
    for plugin in plugins:
        docs_url = plugin.docs or plugin.repo
        if not docs_url:
            continue

        if not printed_empty_line:
            click.echo()
            printed_empty_line = True

        click.echo(
            f"To learn more about {plugin.type.descriptor} '{plugin.name}', visit {docs_url}"
        )



from .analysis_helper import AnalysisHelper
from .base import MeltanoQuery


class SqlUtils:
    def get_sql(self, design, incoming_json, schema=None):
        query = MeltanoQuery(
            definition=incoming_json, design_helper=design, schema=schema
        )

        (sql, query_attributes, aggregate_columns) = query.get_query()

        return {
            "aggregates": aggregate_columns,
            "query_attributes": query_attributes,
            "sql": sql,
        }



"""Defines `meltano elt` command."""
import datetime
import logging
import os
import sys
from contextlib import contextmanager

import click
from async_generator import asynccontextmanager
from meltano.core.db import project_engine
from meltano.core.elt_context import ELTContextBuilder
from meltano.core.job import Job, JobFinder
from meltano.core.job.stale_job_failer import StaleJobFailer
from meltano.core.logging import JobLoggingService, OutputLogger
from meltano.core.plugin import PluginRef, PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.runner import RunnerError
from meltano.core.runner.dbt import DbtRunner
from meltano.core.runner.singer import SingerRunner
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.transform_add_service import TransformAddService
from meltano.core.utils import click_run_async

from . import cli
from .params import pass_project
from .utils import CliError

DUMPABLES = {
    "catalog": (PluginType.EXTRACTORS, "catalog"),
    "state": (PluginType.EXTRACTORS, "state"),
    "extractor-config": (PluginType.EXTRACTORS, "config"),
    "loader-config": (PluginType.LOADERS, "config"),
}

logger = logging.getLogger(__name__)


def logs(*args, **kwargs):
    logger.info(click.style(*args, **kwargs))


@cli.command()
@click.argument("extractor")
@click.argument("loader")
@click.option("--transform", type=click.Choice(["skip", "only", "run"]), default="skip")
@click.option("--dry", help="Do not actually run.", is_flag=True)
@click.option(
    "--full-refresh",
    help="Perform a full refresh (ignore state left behind by any previous runs)",
    is_flag=True,
)
@click.option(
    "--select",
    "-s",
    help="Select only these specific entities for extraction",
    multiple=True,
    default=[],
)
@click.option(
    "--exclude",
    "-e",
    help="Exclude these specific entities from extraction",
    multiple=True,
    default=[],
)
@click.option("--catalog", help="Extractor catalog file")
@click.option("--state", help="Extractor state file")
@click.option(
    "--dump",
    type=click.Choice(DUMPABLES.keys()),
    help="Dump content of pipeline-specific generated file",
)
@click.option(
    "--job_id", envvar="MELTANO_JOB_ID", help="A custom string to identify the job."
)
@click.option(
    "--force",
    "-f",
    help="Force a new run even when a pipeline with the same Job ID is already running",
    is_flag=True,
)
@pass_project(migrate=True)
@click_run_async
async def elt(
    project,
    extractor,
    loader,
    transform,
    dry,
    full_refresh,
    select,
    exclude,
    catalog,
    state,
    dump,
    job_id,
    force,
):
    """
    meltano elt EXTRACTOR_NAME LOADER_NAME

    extractor_name: Which extractor should be used in this extraction
    loader_name: Which loader should be used in this extraction
    """

    select_filter = [*select, *(f"!{entity}" for entity in exclude)]

    job = Job(
        job_id=job_id
        or f'{datetime.datetime.utcnow().strftime("%Y-%m-%dT%H%M%S")}--{extractor}--{loader}'
    )

    _, Session = project_engine(project)
    session = Session()
    try:
        plugins_service = ProjectPluginsService(project)

        context_builder = _elt_context_builder(
            project,
            job,
            session,
            extractor,
            loader,
            transform,
            dry_run=dry,
            full_refresh=full_refresh,
            select_filter=select_filter,
            catalog=catalog,
            state=state,
            plugins_service=plugins_service,
        )

        if dump:
            await dump_file(context_builder, dump)
        else:
            await _run_job(project, job, session, context_builder, force=force)
    finally:
        session.close()

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_elt(extractor=extractor, loader=loader, transform=transform)


def _elt_context_builder(
    project,
    job,
    session,
    extractor,
    loader,
    transform,
    dry_run=False,
    full_refresh=False,
    select_filter=[],
    catalog=None,
    state=None,
    plugins_service=None,
):
    transform_name = None
    if transform != "skip":
        transform_name = _find_transform_for_extractor(
            extractor, plugins_service=plugins_service
        )

    return (
        ELTContextBuilder(project, plugins_service=plugins_service)
        .with_session(session)
        .with_job(job)
        .with_extractor(extractor)
        .with_loader(loader)
        .with_transform(transform_name or transform)
        .with_dry_run(dry_run)
        .with_only_transform(transform == "only")
        .with_full_refresh(full_refresh)
        .with_select_filter(select_filter)
        .with_catalog(catalog)
        .with_state(state)
    )


async def dump_file(context_builder, dumpable):
    """Dump the given dumpable for this pipeline."""
    elt_context = context_builder.context()

    try:
        plugin_type, file_id = DUMPABLES[dumpable]
        invoker = elt_context.invoker_for(plugin_type)

        async with invoker.prepared(elt_context.session):
            content = await invoker.dump(file_id)

        print(content)
    except FileNotFoundError as err:
        raise CliError(f"Could not find {dumpable} file for this pipeline") from err
    except Exception as err:
        raise CliError(f"Could not dump {dumpable}: {err}") from err


async def _run_job(project, job, session, context_builder, force=False):
    StaleJobFailer(job.job_id).fail_stale_jobs(session)

    if not force:
        existing = JobFinder(job.job_id).latest_running(session)
        if existing:
            raise CliError(
                f"Another '{job.job_id}' pipeline is already running which started at {existing.started_at}. "
                + "To ignore this check use the '--force' option."
            )

    async with job.run(session):
        job_logging_service = JobLoggingService(project)
        with job_logging_service.create_log(job.job_id, job.run_id) as log_file:
            output_logger = OutputLogger(log_file)

            await _run_elt(project, context_builder, output_logger)


@asynccontextmanager
async def _redirect_output(output_logger):
    meltano_stdout = output_logger.out("meltano", stream=sys.stdout, color="blue")
    meltano_stderr = output_logger.out("meltano", color="blue")

    with meltano_stdout.redirect_logging(ignore_errors=(CliError,)):
        async with meltano_stdout.redirect_stdout(), meltano_stderr.redirect_stderr():
            try:
                yield
            except CliError as err:
                err.print()
                raise


async def _run_elt(project, context_builder, output_logger):
    async with _redirect_output(output_logger):
        try:
            elt_context = context_builder.context()

            if not elt_context.only_transform:
                await _run_extract_load(elt_context, output_logger)
            else:
                logs("Extract & load skipped.", fg="yellow")

            if elt_context.transformer:
                await _run_transform(elt_context, output_logger)
            else:
                logs("Transformation skipped.", fg="yellow")
        except RunnerError as err:
            raise CliError(f"ELT could not be completed: {err}") from err


async def _run_extract_load(elt_context, output_logger, **kwargs):  # noqa: WPS231
    extractor = elt_context.extractor.name
    loader = elt_context.loader.name

    extractor_log = output_logger.out(extractor, color="yellow")
    loader_log = output_logger.out(loader, color="green")

    @contextmanager
    def nullcontext():
        yield None

    extractor_out_writer = nullcontext
    loader_out_writer = nullcontext
    if logger.getEffectiveLevel() == logging.DEBUG:
        extractor_out = output_logger.out(f"{extractor} (out)", color="bright_yellow")
        loader_out = output_logger.out(f"{loader} (out)", color="bright_green")

        extractor_out_writer = extractor_out.line_writer
        loader_out_writer = loader_out.line_writer

    logs("Running extract & load...")

    singer_runner = SingerRunner(elt_context)
    try:
        with extractor_log.line_writer() as extractor_log_writer, loader_log.line_writer() as loader_log_writer:
            with extractor_out_writer() as extractor_out_writer, loader_out_writer() as loader_out_writer:
                await singer_runner.run(
                    **kwargs,
                    extractor_log=extractor_log_writer,
                    loader_log=loader_log_writer,
                    extractor_out=extractor_out_writer,
                    loader_out=loader_out_writer,
                )
    except RunnerError as err:
        try:
            code = err.exitcodes[PluginType.EXTRACTORS]
            message = extractor_log.last_line.rstrip() or "(see above)"
            logger.error(
                f"{click.style(f'Extraction failed ({code}):', fg='red')} {message}"
            )
        except KeyError:
            pass

        try:
            code = err.exitcodes[PluginType.LOADERS]
            message = loader_log.last_line.rstrip() or "(see above)"
            logger.error(
                f"{click.style(f'Loading failed ({code}):', fg='red')} {message}"
            )
        except KeyError:
            pass

        raise

    logs("Extract & load complete!", fg="green")


async def _run_transform(elt_context, output_logger, **kwargs):
    transformer_log = output_logger.out(elt_context.transformer.name, color="magenta")

    logs("Running transformation...")

    dbt_runner = DbtRunner(elt_context)
    try:
        with transformer_log.line_writer() as transformer_log_writer:
            await dbt_runner.run(**kwargs, log=transformer_log_writer)
    except RunnerError as err:
        try:
            code = err.exitcodes[PluginType.TRANSFORMERS]
            message = transformer_log.last_line.rstrip() or "(see above)"
            logger.error(
                f"{click.style(f'Transformation failed ({code}):', fg='red')} {message}"
            )
        except KeyError:
            pass

        raise

    logs("Transformation complete!", fg="green")


def _find_transform_for_extractor(extractor: str, plugins_service):
    discovery_service = plugins_service.discovery_service
    try:
        extractor_plugin_def = discovery_service.find_definition(
            PluginType.EXTRACTORS, extractor
        )

        # Check if there is a default transform for this extractor
        transform_plugin_def = discovery_service.find_definition_by_namespace(
            PluginType.TRANSFORMS, extractor_plugin_def.namespace
        )

        # Check if the transform has been added to the project
        transform_plugin = plugins_service.get_plugin(transform_plugin_def)

        return transform_plugin.name
    except PluginNotFoundError:
        return None



import click
from meltano.core.plugin import PluginType
from meltano.core.project_add_service import ProjectAddService
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.tracking import GoogleAnalyticsTracker

from . import cli
from .params import pass_project
from .utils import CliError, add_related_plugins, install_plugins


@cli.command()
@click.argument(
    "plugin_type", type=click.Choice(PluginType.cli_arguments()), required=False
)
@click.argument("plugin_name", nargs=-1, required=False)
@click.option("--include-related", is_flag=True)
@click.option(
    "--clean",
    is_flag=True,
    help="Completely reinstall a plugin rather than simply upgrading if necessary.",
)
@click.option(
    "--parallelism",
    "-p",
    type=click.INT,
    default=None,
    help="Limit the number of plugins to install in parallel. Defaults to the number of cores.",
)
@pass_project(migrate=True)
def install(project, plugin_type, plugin_name, include_related, clean, parallelism):
    """
    Installs all the dependencies of your project based on the meltano.yml file.
    Read more at https://www.meltano.com/docs/command-line-interface.html.
    """
    plugins_service = ProjectPluginsService(project)

    if plugin_type:
        plugin_type = PluginType.from_cli_argument(plugin_type)
        plugins = plugins_service.get_plugins_of_type(plugin_type)
        if plugin_name:
            plugins = [p for p in plugins if p.name in plugin_name]
    else:
        plugins = list(plugins_service.plugins())

    if include_related:
        add_service = ProjectAddService(project, plugins_service=plugins_service)
        related_plugins = add_related_plugins(project, plugins, add_service=add_service)
        plugins.extend(related_plugins)

    # We will install the plugins in reverse order, since dependencies
    # are listed after their dependents in `related_plugins`, but should
    # be installed first.
    plugins.reverse()

    click.echo(f"Installing {len(plugins)} plugins...")

    success = install_plugins(project, plugins, parallelism=parallelism, clean=clean)

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_install()

    if not success:
        raise CliError("Failed to install plugin(s)")



import logging

import click
from flask_security.utils import hash_password
from meltano.api.app import create_app
from meltano.core.utils import identity

from . import cli
from .params import pass_project


@cli.group(invoke_without_command=True)
@pass_project(migrate=True)
@click.pass_context
def user(ctx, project):
    ctx.obj["project"] = project


@user.command()
@click.argument("username")
@click.argument("password")
@click.option("--overwrite", "-f", is_flag=True, default=False)
@click.option("--role", "-G", multiple=True)
@click.pass_context
def add(ctx, username, password, role=[], **flags):
    app = create_app()

    from meltano.api.security import users

    try:
        with app.app_context():
            # make sure our User doesn't already exist
            if not flags["overwrite"] and users.get_user(username):
                raise Exception(
                    f"User '{username}' already exists. Use --overwrite to update it."
                )

            # make sure all roles exists
            roles = []
            for role_name in role:
                r = users.find_role(role_name)
                if not r:
                    raise Exception(f"Role '{role_name}' does not exists.")

                roles.append(r)

            current_user = users.get_user(username) or users.create_user(
                username=username
            )
            current_user.password = hash_password(password)
            current_user.roles = roles

            # for some reason the scoped_session doesn't trigger the commit
            users.db.session.commit()
    except Exception as err:
        click.secho(f"Could not create user '{username}': {err}", fg="red")
        click.Abort()



"""Install plugins into the project, using pip in separate virtual environments by default."""
import asyncio
import functools
import logging
import sys
from enum import Enum
from multiprocessing import cpu_count
from typing import Any, Callable, Iterable, Tuple

from meltano.core.plugin.project_plugin import ProjectPlugin

from .compiler.project_compiler import ProjectCompiler
from .error import AsyncSubprocessError, PluginInstallError, PluginInstallWarning
from .plugin import PluginType
from .plugin_discovery_service import PluginDiscoveryService
from .project import Project
from .project_add_service import ProjectAddService
from .project_plugins_service import ProjectPluginsService
from .utils import noop, run_async
from .venv_service import VenvService


class PluginInstallReason(str, Enum):
    ADD = "add"
    INSTALL = "install"
    UPGRADE = "upgrade"


class PluginInstallStatus(Enum):
    """The status of the process of installing a plugin."""

    RUNNING = "running"
    SUCCESS = "success"
    SKIPPED = "skipped"
    ERROR = "error"
    WARNING = "warning"


class PluginInstallState:
    """A message reporting the progress of installing a plugin."""

    def __init__(
        self,
        plugin: ProjectPlugin,
        reason: PluginInstallReason,
        status: PluginInstallStatus,
        message: str = None,
        details: str = None,
    ):
        # TODO: use dataclasses.dataclass for this when 3.6 support is dropped
        self.plugin = plugin
        self.reason = reason
        self.status = status
        self.message = message
        self.details = details

    @property
    def successful(self):
        """If the installation completed without error."""
        return self.status in {PluginInstallStatus.SUCCESS, PluginInstallStatus.SKIPPED}

    @property
    def skipped(self):
        """Return 'True' if the installation was skipped / not needed."""
        return self.status == PluginInstallStatus.SKIPPED

    @property  # noqa: WPS212  # Too many return statements
    def verb(self):
        """Verb form of status."""
        if self.status is PluginInstallStatus.RUNNING:
            if self.reason is PluginInstallReason.UPGRADE:
                return "Updating"
            return "Installing"

        if self.status is PluginInstallStatus.SUCCESS:
            if self.reason is PluginInstallReason.UPGRADE:
                return "Updated"

            return "Installed"

        if self.status is PluginInstallStatus.SKIPPED:
            return "Skipped installing"

        return "Errored"


def installer_factory(project, plugin: ProjectPlugin, *args, **kwargs):
    cls = PipPluginInstaller

    if hasattr(plugin, "installer_class"):
        cls = plugin.installer_class

    return cls(project, plugin, *args, **kwargs)


def with_semaphore(func):
    """Gate access to the method using its class's semaphore."""

    @functools.wraps(func)  # noqa: WPS430
    async def wrapper(self, *args, **kwargs):  # noqa: WPS430
        async with self.semaphore:
            return await func(self, *args, **kwargs)

    return wrapper


class PluginInstallService:
    def __init__(
        self,
        project: Project,
        plugins_service: ProjectPluginsService = None,
        status_cb: Callable[[PluginInstallState], Any] = noop,
        parallelism=None,
        clean=False,
    ):
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)
        self.status_cb = status_cb

        if parallelism is None:
            parallelism = cpu_count()
        if parallelism < 1:
            parallelism = sys.maxsize  # unbounded
        self.semaphore = asyncio.Semaphore(parallelism)
        self.clean = clean

    def install_all_plugins(
        self, reason=PluginInstallReason.INSTALL
    ) -> Tuple[PluginInstallState]:
        """
        Install all the plugins for the project.

        Blocks until all plugins are installed.
        """
        return self.install_plugins(self.plugins_service.plugins(), reason=reason)

    def install_plugins(
        self,
        plugins: Iterable[ProjectPlugin],
        reason=PluginInstallReason.INSTALL,
    ) -> Tuple[PluginInstallState]:
        """
        Install all the provided plugins.

        Blocks until all plugins are installed.
        """
        return run_async(self.install_plugins_async(plugins, reason=reason))

    async def install_plugins_async(
        self,
        plugins: Iterable[ProjectPlugin],
        reason=PluginInstallReason.INSTALL,
    ) -> Tuple[PluginInstallState]:
        """Install all the provided plugins."""
        results = await asyncio.gather(
            *[
                self.install_plugin_async(plugin, reason, compile_models=False)
                for plugin in plugins
            ]
        )
        for plugin in plugins:
            if plugin.type is PluginType.MODELS:
                self.compile_models()
                break

        return results

    def install_plugin(
        self,
        plugin: ProjectPlugin,
        reason=PluginInstallReason.INSTALL,
        compile_models=True,
    ) -> PluginInstallState:
        """
        Install a plugin.

        Blocks until the plugin is installed.
        """
        return run_async(
            self.install_plugin_async(
                plugin,
                reason=reason,
                compile_models=compile_models,
            )
        )

    @with_semaphore
    async def install_plugin_async(
        self,
        plugin: ProjectPlugin,
        reason=PluginInstallReason.INSTALL,
        compile_models=True,
    ) -> PluginInstallState:
        """Install a plugin."""
        self.status_cb(
            PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.RUNNING,
            )
        )
        if not plugin.is_installable():
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.SKIPPED,
                message=f"Plugin '{plugin.name}' does not require installation",
            )
            self.status_cb(state)
            return state

        try:
            async with plugin.trigger_hooks("install", self, plugin, reason):
                await installer_factory(self.project, plugin).install(
                    reason, self.clean
                )
                state = PluginInstallState(
                    plugin=plugin, reason=reason, status=PluginInstallStatus.SUCCESS
                )
                if compile_models and plugin.type is PluginType.MODELS:
                    self.compile_models()
                self.status_cb(state)
                return state

        except PluginInstallError as err:
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.ERROR,
                message=str(err),
            )
            self.status_cb(state)
            return state

        except PluginInstallWarning as warn:
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.WARNING,
                message=str(warn),
            )
            self.status_cb(state)
            return state

        except AsyncSubprocessError as err:
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.ERROR,
                message=f"{plugin.type.descriptor} '{plugin.name}' could not be installed: {err}".capitalize(),
                details=await err.stderr,
            )
            self.status_cb(state)
            return state

    def compile_models(self):
        compiler = ProjectCompiler(self.project)
        try:
            compiler.compile()
        except Exception:
            pass


class PipPluginInstaller:
    def __init__(
        self,
        project,
        plugin: ProjectPlugin,
        venv_service: VenvService = None,
    ):
        self.plugin = plugin
        self.venv_service = venv_service or VenvService(
            project,
            namespace=self.plugin.type,
            name=self.plugin.name,
        )

    async def install(self, reason, clean):
        """Install the plugin into the virtual environment using pip."""
        return await self.venv_service.install(
            self.plugin.formatted_pip_url, clean=clean
        )



import glob
import json
import os
import uuid
from copy import deepcopy
from datetime import datetime, timezone
from pathlib import Path
from typing import Dict, List

import networkx as nx
import sqlparse
from meltano.core.plugin import PluginType
from meltano.core.plugin.model import Package
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.project import Project
from meltano.core.sql.base import (
    MeltanoAggregate,
    MeltanoColumn,
    MeltanoTable,
    MeltanoTimeframe,
    MeltanoTimeframePeriod,
)
from meltano.core.sql.design_helper import PypikaJoinExecutor
from meltano.core.utils import NotFound, find_named, slugify
from networkx.readwrite import json_graph
from pyhocon import ConfigFactory


class MeltanoAnalysisFileParserError(Exception):
    def __init__(self, message, file_name):
        self.message = message
        self.file_name = file_name
        super().__init__(f"{file_name}: {message}")


class MeltanoAnalysisFileParserMissingTableError(MeltanoAnalysisFileParserError):
    def __init__(self, field, your_choice, cls, file_name):
        message = (
            f'Missing accompanying table "{your_choice}" in "{field}" field in {cls}'
        )
        super().__init__(message, file_name)


class MeltanoAnalysisMissingTopicFilesError(MeltanoAnalysisFileParserError):
    def __init__(self, message):
        super().__init__(message, "topic file")


class MeltanoAnalysisFileParserUnacceptableChoiceError(MeltanoAnalysisFileParserError):
    def __init__(self, field, your_choice, acceptable_choices, cls, file_name):
        quoted_choices = [f'"{a}"' for a in acceptable_choices]
        message = f"Unacceptable choice for field {field}. You wrote \"{your_choice}\". Acceptable choices are: {', '.join(quoted_choices)}"
        super().__init__(message, file_name)


class MeltanoAnalysisFileParserMissingFieldsError(MeltanoAnalysisFileParserError):
    def __init__(self, fields, cls, file_name):
        field = "field" if len(fields) == 1 else "fields"
        message = f"Missing {field} in {cls}: \"{', '.join(fields)}\""
        super().__init__(message, file_name)


class MeltanoAnalysisInvalidJoinDependencyError(MeltanoAnalysisFileParserError):
    def __init__(self, field, cls, file_name):
        message = f'Invalid join dependency for join {cls} in in field "{field}"'
        super().__init__(message, file_name)


class MeltanoAnalysisFileParser:
    def __init__(self, project: Project):
        self.project = project
        self._output_dir = self.project.run_dir("models")
        self.topics = []
        self.tables = []
        self.packaged_topics = []
        self.packaged_tables = []
        self.required_topic_properties = ["name", "label", "designs"]
        self.required_design_properties = ["from", "label", "description"]
        self.required_join_properties = ["sql_on", "relationship"]
        self.required_table_properties = ["sql_table_name", "columns"]
        self.join_relationship_types = [
            "one_to_one",
            "one_to_many",
            "many_to_one",
            "many_to_many",
        ]

    def uses_accepted_choice(self, choices, choice):
        return choice in choices

    def missing_properties(self, properties, properties_dict):
        properties_copy = properties.copy()
        for prop in properties_dict:
            try:
                property_index = properties_copy.index(prop)
                del properties_copy[property_index]
            except ValueError as e:
                continue
        return properties_copy

    def parse_m5o_file(self, file_path):
        try:
            parsed = ConfigFactory.parse_file(file_path)
            parsed["_file_path"] = str(file_path)
            return parsed
        except Exception as e:
            raise MeltanoAnalysisFileParserError(str(e), str(file_path.parts[-1]))

    def generate_join_graph_for_node(self, graph, design, source_name, joins):
        remaining_joins = []
        check_against = []

        for join in joins:
            join_executor = None
            if "executor" in join:
                join_executor = join["executor"]
            else:
                join_executor = join["executor"] = PypikaJoinExecutor(design, join)
                join_executor.visit(sqlparse.parse(join["sql_on"])[0])

            # the `table` identifier here represent either:
            #   - Design name
            #   - Join name
            left_join_candidate = join_executor.comparison_fields.left.table
            right_join_candidate = join_executor.comparison_fields.right.table

            if left_join_candidate == source_name:
                check_against.append(right_join_candidate)

                # that means we must add the `right` join
                graph.add_node(right_join_candidate)
                graph.add_edge(left_join_candidate, right_join_candidate)
            elif right_join_candidate == source_name:
                check_against.append(left_join_candidate)

                # that means we must add the `left` join
                graph.add_node(left_join_candidate)
                graph.add_edge(right_join_candidate, left_join_candidate)
            else:
                # both `left` & `right` are joins
                remaining_joins.append(join)

        # recurse through the remaining joins
        for check in check_against:
            self.generate_join_graph_for_node(graph, design, check, remaining_joins)

    def graph_design(self, design):
        # The `base` table should always be part of the graph
        # base_table = design["related_table"]["name"]
        source_name = design["name"]

        graph = nx.DiGraph()
        graph.add_node(source_name)
        joins = deepcopy(design.get("joins", []))

        self.generate_join_graph_for_node(graph, design, source_name, joins)
        return json_graph.node_link_data(graph)

    def graph_topic(self, topic):
        for design in topic["designs"]:
            design_graph = self.graph_design(design)
            design["graph"] = design_graph

        return topic

    def compile(self, topics):
        indices = {}
        for topic in topics:
            # Each topic is:
            # - indexed using its topic['namespace']
            #   e.g. package.name for a package
            #   `cutom/dir1/dir2` for a custom topic defined under `project/models/dir1/dir2/`
            # - compiled and stored under `.meltano/run/models/{topic['namespace']}`
            #   e.g. `.meltano/run/models/custom/dir1/dir2/mytopic.topic.m5oc`
            topic_identifier = Path(topic["namespace"]).joinpath(topic["name"])

            topic = self.graph_topic(topic)
            indices[str(topic_identifier)] = {
                "plugin_namespace": topic["plugin_namespace"],
                "namespace": topic["namespace"],
                "name": topic["name"],
                "designs": [e["name"] for e in topic["designs"]],
            }

            # Recursively create the missing directories under `.meltano/run/`
            compiled_file_dir = self._output_dir.joinpath(topic["namespace"])
            compiled_file_dir.mkdir(parents=True, exist_ok=True)

            # Create the m5oc file for the topic
            compiled_file_name = f"{topic['name']}.topic.m5oc"
            compiled_file_path = compiled_file_dir.joinpath(compiled_file_name)
            with compiled_file_path.open("w") as compiled_topic:
                json.dump(topic, compiled_topic)

        # Add the final index file under `.meltano/run/models/`
        index_file_path = self._output_dir.joinpath("topics.index.m5oc")
        with index_file_path.open("w") as index_file:
            index_file.write(json.dumps(indices))

    def parse_packages(self):
        discovery = PluginDiscoveryService(self.project)
        model_plugins = discovery.get_plugins_of_type(PluginType.MODELS)

        for package in self.packages():
            if not package.topics and package.tables:
                raise MeltanoAnalysisMissingTopicFilesError(
                    f"Missing topic file(s) for package {package}"
                )

            for table in package.tables:
                conf = self.parse_m5o_file(table)
                parsed_table = self.table(conf, table.name)
                self.tables.append(parsed_table)

            for topic in package.topics:
                conf = self.parse_m5o_file(topic)
                parsed_topic = self.topic(conf, topic.name)
                model = next(
                    (plugin for plugin in model_plugins if plugin.name == package.name),
                    None,
                )
                parsed_topic["plugin_namespace"] = model.namespace if model else None
                parsed_topic["namespace"] = package.name
                self.packaged_topics.append(parsed_topic)

            # Reset the tables list so that tables with the same name from
            #  different packages are not interfearing with correctly parsing
            #  the packages that follow
            self.tables = []

        return self.packaged_topics

    def parse(self):
        # Iterate through the project/models/ directory and all its subdirectories
        # Each directory sets a namespace for the topics and tables that exist
        #  inside it and all the tables and topics in each namespace are parsed
        #  together (so we assume unique names and they can reference each other).
        models = self.project.root_dir("model")
        subfolders = [f for f in models.glob("**/*") if f.is_dir()]

        for folder in [models] + subfolders:
            self.m5o_tables = list(folder.glob("*.table.m5o"))
            self.m5o_topics = list(folder.glob("*.topic.m5o"))
            if not self.m5o_topics and self.m5o_tables:
                raise MeltanoAnalysisMissingTopicFilesError("Missing topic file(s)")

            for table in self.m5o_tables:
                conf = self.parse_m5o_file(table)
                parsed_table = self.table(conf, table.name)
                self.tables.append(parsed_table)

            for topic in self.m5o_topics:
                conf = self.parse_m5o_file(topic)
                parsed_topic = self.topic(conf, topic.name)

                # The namespace for a topic in a directory under project/models/
                #   is `custom/path/to/directory`
                # For example:
                # `project/models/mytopic.topic.m5o` --> `custom`
                # `project/models/dir1/dir2/mytopic.topic.m5o` --> `custom/dir1/dir2`
                parsed_topic["namespace"] = str(
                    Path("custom").joinpath(folder.relative_to(models))
                )
                parsed_topic["plugin_namespace"] = conf.get(
                    "plugin_namespace", folder.name
                )

                self.topics.append(parsed_topic)

            self.tables = []

        return self.topics

    def topic(self, ma_file_topic_dict, file_name):
        temp_topic = {}
        missing_properties = self.missing_properties(
            self.required_topic_properties, ma_file_topic_dict
        )

        if missing_properties:
            raise MeltanoAnalysisFileParserMissingFieldsError(
                missing_properties, "topic", file_name
            )

        for prop_name, prop_def in ma_file_topic_dict.items():
            temp_topic[prop_name] = prop_def
            if prop_name == "designs":
                temp_topic[prop_name] = self.designs(prop_def, file_name)

        return temp_topic

    def table_conf_by_name(self, table_name, cls, prop, file_name):
        try:
            return find_named(self.tables, table_name)
        except NotFound as e:
            raise MeltanoAnalysisFileParserMissingTableError(
                prop, table_name, cls, file_name
            )

    def add_keys_to_table_def(self, table_def, source_name):
        mtable = MeltanoTable(table_def)
        mtable.source_name = source_name

        for column_def in table_def["columns"]:
            mcolumn = MeltanoColumn(definition=column_def, table=mtable)
            column_def["key"] = mcolumn.alias()

        for aggregate_def in table_def["aggregates"]:
            maggregate = MeltanoAggregate(definition=aggregate_def, table=mtable)
            aggregate_def["key"] = maggregate.alias()

        for timeframe_def in table_def["timeframes"]:
            mtimeframe = MeltanoTimeframe(definition=timeframe_def, table=mtable)
            timeframe_def["key"] = mtimeframe.alias()
            for period_def in timeframe_def.get("periods", []):
                mperiod = MeltanoTimeframePeriod(
                    definition=period_def, timeframe=mtimeframe
                )
                period_def["key"] = mperiod.alias()

    def designs(self, ma_file_designs_dict, file_name):
        topic_designs = []
        for design_name, design_def in ma_file_designs_dict.items():
            temp_design = {}
            temp_design["name"] = design_name
            missing_properties = self.missing_properties(
                self.required_design_properties, design_def
            )
            if missing_properties:
                raise MeltanoAnalysisFileParserMissingFieldsError(
                    missing_properties, "design", file_name
                )

            for prop_name, prop_def in design_def.items():
                temp_design[prop_name] = prop_def
                if prop_name == "from":
                    related_table = deepcopy(
                        self.table_conf_by_name(
                            temp_design[prop_name], "design", prop_name, file_name
                        )
                    )

                    self.add_keys_to_table_def(related_table, design_name)

                    temp_design["related_table"] = related_table
                if prop_name == "joins":
                    temp_design[prop_name] = self.joins(prop_def, file_name)

            topic_designs.append(temp_design)

        return topic_designs

    def joins(self, ma_file_joins_dict, file_name):
        design_joins = []
        for join_name, join_def in ma_file_joins_dict.items():
            temp_join = {}
            temp_join["name"] = join_name

            related_table_name = join_def.get("from", join_name)
            related_table = deepcopy(
                self.table_conf_by_name(related_table_name, "join", "name", file_name)
            )

            self.add_keys_to_table_def(related_table, join_name)

            temp_join["related_table"] = related_table

            missing_properties = self.missing_properties(
                self.required_join_properties, join_def
            )
            if missing_properties:
                raise MeltanoAnalysisFileParserMissingFieldsError(
                    missing_properties, "join", file_name
                )

            for prop_name, prop_def in join_def.items():
                temp_join[prop_name] = prop_def
                if prop_name == "relationship":
                    uses_accepted_choices = self.uses_accepted_choice(
                        self.join_relationship_types, prop_def
                    )
                    if not uses_accepted_choices:
                        raise MeltanoAnalysisFileParserUnacceptableChoiceError(
                            prop_name,
                            prop_def,
                            self.join_relationship_types,
                            "join",
                            file_name,
                        )

            design_joins.append(temp_join)

        return design_joins

    def table(self, table_file, file_name):
        temp_table = {}
        missing_properties = self.missing_properties(
            self.required_table_properties, table_file
        )
        if missing_properties:
            raise MeltanoAnalysisFileParserMissingFieldsError(
                missing_properties, "table", file_name
            )

        # set all the subproperties
        for prop_name in ("columns", "aggregates", "timeframes"):
            prop_def = table_file.get(prop_name, {})
            temp_table[prop_name] = self.name_flatten_dict(prop_def)

        # set the rest of the properties
        for prop_name, prop_def in table_file.items():
            # subproperties are already set
            if prop_name in temp_table:
                continue

            temp_table[prop_name] = prop_def

        return temp_table

    @classmethod
    def name_flatten_dict(cls, d: Dict) -> List:
        return [{"name": k, **rest} for k, rest in d.items()]

    @staticmethod
    def fill_base_m5o_dict(file, name, file_dict=None):
        if file_dict is None:
            file_dict = {"name": name}

        file_dict["path"] = str(file)
        file_dict["slug"] = slugify(name)

        if "id" not in file_dict:
            file_dict["id"] = str(uuid.uuid4())

        # Legacy reports and dashboards can have a `createdAt` key that would
        # conflict with the new `created_at` key instead of being overwritten.
        file_dict.pop("createdAt", None)

        file_dict["created_at"] = datetime.now(timezone.utc).timestamp()

        return file_dict

    def packages(self) -> List[Package]:
        return [
            Package(f.name, self.project)
            for f in self.project.model_dir().iterdir()
            if f.is_dir()
        ]

    @classmethod
    def package_files(cls, package):
        m5oFiles = {
            "topics": {"label": "Topics", "items": []},
            "tables": {"label": "Tables", "items": []},
        }
        for f in package.files:
            basename = os.path.basename(f)
            filename, ext = os.path.splitext(basename)
            if ext != ".m5o":
                continue

            # filename splittext twice occurs due to current *.type.extension convention (two dots)
            filename = f"{package.name}/{filename.lower()}"
            filename, ext = os.path.splitext(filename)
            file_dict = cls.fill_base_m5o_dict(f, filename)
            if ext == ".topic":
                m5oFiles["topics"]["items"].append(file_dict)
            if ext == ".table":
                m5oFiles["tables"]["items"].append(file_dict)

        return m5oFiles



import logging
import os
from collections import OrderedDict
from pathlib import Path

import sqlalchemy
from flask import jsonify, redirect, url_for
from meltano.api.json import freeze_keys
from meltano.api.models import db
from meltano.api.security import create_dev_user
from meltano.core.connection_service import ConnectionService, DialectNotSupportedError
from meltano.core.elt_context import ELTContextBuilder
from meltano.core.m5o.m5oc_file import M5ocFile
from meltano.core.project import Project
from meltano.core.sql.analysis_helper import AnalysisHelper
from meltano.core.sql.sql_utils import SqlUtils
from pypika import Order, Query
from sqlalchemy.event import listen

from .settings_helper import SettingsHelper

ENABLED_DIALECTS = ["postgres", "sqlite"]


class ConnectionNotFound(Exception):
    def __init__(self, connection_name: str):
        self.connection_name = connection_name
        super().__init__(f"{connection_name} is missing.")


class UnsupportedConnectionDialect(Exception):
    def __init__(self, connection_dialect: str):
        self.connection_dialect = connection_dialect
        super().__init__(f"Dialect {connection_dialect} is not supported.")


class SqlHelper(SqlUtils):
    def parse_sql(self, input):
        placeholders = self.placeholder_match(input)

    def placeholder_match(self, input):
        outer_pattern = r"(\$\{[\w\.]*\})"
        inner_pattern = r"\$\{([\w\.]*)\}"
        outer_results = re.findall(outer_pattern, input)
        inner_results = re.findall(inner_pattern, input)
        return (outer_results, inner_results)

    def get_m5oc_topic(self, namespace, topic_name):
        project = Project.find()
        m5oc_file = project.run_dir("models", namespace, f"{topic_name}.topic.m5oc")
        return M5ocFile.load(m5oc_file)

    def get_db_engine(self, extractor, loader, transform):
        project = Project.find()
        context = (
            ELTContextBuilder(project)
            .with_session(db.session)
            .with_extractor(extractor)
            .with_loader(loader)
            .with_transform(transform)
            .context()
        )
        connection_service = ConnectionService(context)

        engine_hooks = []
        dialect = connection_service.dialect
        engine_uri = None
        params = None

        try:
            params = connection_service.analyze_params()
            engine_uri = connection_service.analyze_uri()

            if dialect not in ENABLED_DIALECTS:
                raise UnsupportedConnectionDialect(dialect)
        except DialectNotSupportedError:
            raise UnsupportedConnectionDialect(dialect)

        if dialect == "postgres":

            def set_connection_schema(raw, conn):
                schema = params["schema"]
                with raw.cursor() as cursor:
                    res = cursor.execute(f"SET search_path TO {schema};")
                    logging.debug(f"Connection schema set to {schema}")

            engine_hooks.append(
                lambda engine: listen(engine, "first_connect", set_connection_schema)
            )

        engine = sqlalchemy.create_engine(engine_uri)
        for hook in engine_hooks:
            hook(engine)

        return engine

    # we need to `freeze` each result to make sure
    # the attribute name will be correct for the lookup
    def get_query_results(self, extractor, loader, transform, sql):
        engine = self.get_db_engine(extractor, loader, transform)
        results = engine.execute(sqlalchemy.text(sql))
        results = [freeze_keys(OrderedDict(row)) for row in results]
        return results

    def reset_db(self):
        try:
            db.drop_all()
        except sqlalchemy.exc.OperationalError as err:
            logging.error("Failed drop database.")

        db.create_all()
        create_dev_user()



import json
import tempfile
from pathlib import Path

import click
import dotenv
from meltano.core.db import project_engine
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.settings_service import SettingValueStore, StoreNotSupportedError

from . import cli
from .params import pass_project
from .utils import CliError


@cli.group(invoke_without_command=True)
@click.option(
    "--plugin-type", type=click.Choice(PluginType.cli_arguments()), default=None
)
@click.argument("plugin_name")
@click.option("--format", type=click.Choice(["json", "env"]), default="json")
@click.option("--extras", is_flag=True)
@pass_project(migrate=True)
@click.pass_context
def config(ctx, project, plugin_type, plugin_name, format, extras):
    plugin_type = PluginType.from_cli_argument(plugin_type) if plugin_type else None

    plugins_service = ProjectPluginsService(project)

    try:
        plugin = plugins_service.find_plugin(
            plugin_name, plugin_type=plugin_type, configurable=True
        )
    except PluginNotFoundError:
        if plugin_name == "meltano":
            plugin = None
        else:
            raise

    _, Session = project_engine(project)
    session = Session()
    try:
        if plugin:
            settings = PluginSettingsService(
                project, plugin, plugins_service=plugins_service
            )
        else:
            settings = ProjectSettingsService(
                project, config_service=plugins_service.config_service
            )

        ctx.obj["settings"] = settings
        ctx.obj["session"] = session

        if ctx.invoked_subcommand is None:
            if format == "json":
                process = extras is not True
                config = settings.as_dict(
                    extras=extras, process=process, session=session
                )
                print(json.dumps(config, indent=2))
            elif format == "env":
                env = settings.as_env(extras=extras, session=session)

                with tempfile.NamedTemporaryFile() as temp_dotenv:
                    path = temp_dotenv.name
                    for key, value in env.items():
                        dotenv.set_key(path, key, value)

                    dotenv_content = Path(temp_dotenv.name).read_text()

                print(dotenv_content, end="")
    finally:
        session.close()


@config.command()
@click.argument("setting_name", nargs=-1, required=True)
@click.argument("value")
@click.option(
    "--store",
    type=click.Choice(SettingValueStore.writables()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
def set(ctx, setting_name, value, store):
    store = SettingValueStore(store)

    try:
        value = json.loads(value)
    except json.JSONDecodeError:
        pass

    settings = ctx.obj["settings"]
    session = ctx.obj["session"]

    path = list(setting_name)
    try:
        value, metadata = settings.set_with_metadata(
            path, value, store=store, session=session
        )
    except StoreNotSupportedError as err:
        raise CliError(
            f"{settings.label.capitalize()} setting '{path}' could not be set in {store.label}: {err}"
        ) from err

    name = metadata["name"]
    store = metadata["store"]
    click.secho(
        f"{settings.label.capitalize()} setting '{name}' was set in {store.label}: {value!r}",
        fg="green",
    )

    current_value, source = settings.get_with_source(name, session=session)
    if source != store:
        click.secho(
            f"Current value is still: {current_value!r} (from {source.label})",
            fg="yellow",
        )


@config.command()
@click.argument("setting_name", nargs=-1, required=True)
@click.option(
    "--store",
    type=click.Choice(SettingValueStore.writables()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
def unset(ctx, setting_name, store):
    store = SettingValueStore(store)

    settings = ctx.obj["settings"]
    session = ctx.obj["session"]

    path = list(setting_name)
    try:
        metadata = settings.unset(path, store=store, session=session)
    except StoreNotSupportedError as err:
        raise CliError(
            f"{settings.label.capitalize()} setting '{path}' in {store.label} could not be unset: {err}"
        ) from err

    name = metadata["name"]
    store = metadata["store"]
    click.secho(
        f"{settings.label.capitalize()} setting '{name}' in {store.label} was unset",
        fg="green",
    )

    current_value, source = settings.get_with_source(name, session=session)
    if source is not SettingValueStore.DEFAULT:
        click.secho(
            f"Current value is now: {current_value!r} (from {source.label})",
            fg="yellow",
        )


@config.command()
@click.option(
    "--store",
    type=click.Choice(SettingValueStore.writables()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
def reset(ctx, store):
    store = SettingValueStore(store)

    settings = ctx.obj["settings"]
    session = ctx.obj["session"]

    try:
        metadata = settings.reset(store=store, session=session)
    except StoreNotSupportedError as err:
        raise CliError(
            f"{settings.label.capitalize()} settings in {store.label} could not be reset: {err}"
        ) from err

    store = metadata["store"]
    click.secho(
        f"{settings.label.capitalize()} settings in {store.label} were reset",
        fg="green",
    )


@config.command("list")
@click.option("--extras", is_flag=True)
@click.pass_context
def list_settings(ctx, extras):
    settings = ctx.obj["settings"]
    session = ctx.obj["session"]

    printed_custom_heading = False
    printed_extra_heading = extras

    # If `--extras` is not specified (`False`), we still want to load both
    # regular and extra settings, since we show custom extras.
    load_extras = True if extras else None

    full_config = settings.config_with_metadata(session=session, extras=load_extras)
    for name, config_metadata in full_config.items():
        value = config_metadata["value"]
        source = config_metadata["source"]
        setting_def = config_metadata["setting"]

        if extras:
            if not setting_def.is_extra:
                continue

            if setting_def.is_custom and not printed_custom_heading:
                click.echo()
                click.echo("Custom:")
                printed_custom_heading = True
        elif setting_def.is_extra:
            if not setting_def.is_custom:
                continue

            if not printed_extra_heading:
                click.echo()
                click.echo("Custom extras, plugin-specific options handled by Meltano:")
                printed_extra_heading = True
        elif setting_def.is_custom and not printed_custom_heading:
            click.echo()
            click.echo("Custom, possibly unsupported by the plugin:")
            printed_custom_heading = True

        click.secho(name, fg="blue", nl=False)

        env_keys = [var.definition for var in settings.setting_env_vars(setting_def)]
        click.echo(f" [env: {', '.join(env_keys)}]", nl=False)

        if source is not SettingValueStore.DEFAULT:
            default_value = setting_def.value
            if default_value is not None:
                click.echo(f" (default: {default_value!r})", nl=False)

        if source is SettingValueStore.DEFAULT:
            label = "default"
        elif source is SettingValueStore.INHERITED:
            label = f"inherited from '{settings.plugin.parent.name}'"
        else:
            label = f"from {source.label}"

        current_value = click.style(f"{value!r}", fg="green")
        click.echo(f" current value: {current_value}", nl=False)

        unexpanded_value = config_metadata.get("unexpanded_value")
        if not unexpanded_value or unexpanded_value == value:
            click.echo(f" ({label})")
        else:
            click.echo(f" ({label}: {unexpanded_value!r})")

        if setting_def.description:
            click.echo("\t", nl=False)
            if setting_def.label:
                click.echo(f"{setting_def.label}: ", nl=False)
            click.echo(f"{setting_def.description}")

    docs_url = settings.docs_url
    if docs_url:
        click.echo()
        click.echo(
            f"To learn more about {settings.label} and its settings, visit {docs_url}"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['meltano',
 'meltano.api',
 'meltano.api.controllers',
 'meltano.api.events',
 'meltano.api.executor',
 'meltano.api.models',
 'meltano.api.security',
 'meltano.api.workers',
 'meltano.cli',
 'meltano.core',
 'meltano.core.behavior',
 'meltano.core.bundle',
 'meltano.core.compiler',
 'meltano.core.job',
 'meltano.core.logging',
 'meltano.core.m5o',
 'meltano.core.plugin',
 'meltano.core.plugin.dbt',
 'meltano.core.plugin.model',
 'meltano.core.plugin.singer',
 'meltano.core.runner',
 'meltano.core.sql',
 'meltano.core.tracking',
 'meltano.core.utils',
 'meltano.migrations',
 'meltano.migrations.versions',
 'meltano.oauth']

package_data = \
{'': ['*'],
 'meltano.api': ['static/*',
                 'static/css/*',
                 'static/js/*',
                 'static/logos/*',
                 'templates/*',
                 'templates/email/pipeline_manual_run/*',
                 'templates/security/*',
                 'templates/security/email/*'],
 'meltano.oauth': ['templates/*']}

install_requires = \
['aiohttp>=3.4.4,<4.0.0',
 'alembic>=1.5.0,<2.0.0',
 'async_generator>=1.10,<2.0',
 'atomicwrites>=1.2.1,<2.0.0',
 'authlib>=0.10,<0.11',
 'bcrypt>=3.2.0,<4.0.0',
 'click-default-group>=1.2.1,<2.0.0',
 'click>=7.0,<8.0',
 'email-validator>=1.1.2,<2.0.0',
 'fasteners>=0.15.0,<0.16.0',
 'flask-cors>=3.0.7,<4.0.0',
 'flask-executor>=0.9.2,<0.10.0',
 'flask-restful>=0.3.7,<0.4.0',
 'flask-sqlalchemy>=2.4.4,<3.0.0',
 'flask>=1.1.2,<2.0.0',
 'flatten-dict>=0.1.0,<0.2.0',
 'gunicorn>=19.9.0,<20.0.0',
 'ipython>=7.5.0,<8.0.0',
 'jsonschema>=2.6.0,<3.0.0',
 'markdown>=3.0.1,<4.0.0',
 'meltano-flask-security>=0.1.0,<0.2.0',
 'networkx>=2.2,<3.0',
 'psutil>=5.6.3,<6.0.0',
 'psycopg2-binary>=2.8.5,<3.0.0',
 'pyhocon>=0.3.51,<0.4.0',
 'pyhumps==1.2.2',
 'pypika>=0.25.1,<0.26.0',
 'python-dotenv>=0.14.0,<0.15.0',
 'python-gitlab>=1.8.0,<2.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.23.0,<3.0.0',
 'simplejson>=3.16.0,<4.0.0',
 'smtpapi>=0.4.1,<0.5.0',
 'snowflake-sqlalchemy>=1.2.3,<2.0.0',
 'sqlalchemy>=1.3.19,<2.0.0',
 'sqlparse>=0.3.0,<0.4.0',
 'watchdog>=0.9.0,<0.10.0',
 'werkzeug>=0.16.1,<0.17.0']

entry_points = \
{'console_scripts': ['meltano = meltano.cli:main']}

setup_kwargs = {
    'name': 'meltano',
    'version': '1.80.1',
    'description': 'Meltano',
    'long_description': '# Meltano: ELT for the DataOps era\n\n[![pipeline status](https://gitlab.com/meltano/meltano/badges/master/pipeline.svg)](https://gitlab.com/meltano/meltano/commits/master)\n\n[Meltano](https://meltano.com) is\n[open source](https://gitlab.com/meltano/meltano),\n[self-hosted](https://meltano.com/docs/production.html),\n[CLI-first](https://meltano.com/docs/command-line-interface.html),\n[debuggable](https://meltano.com/docs/command-line-interface.html#debugging), and\n[extensible](https://meltano.com/docs/plugins.html).\n\n[Pipelines are code](https://meltano.com/docs/project.html),\nready to be version controlled,\n[containerized](https://meltano.com/docs/containerization.html), and\n[deployed continuously](https://meltano.com/docs/production.html#and-onto-the-production-environment).\nDevelop and test\n[locally](https://meltano.com/docs/getting-started.html#local-installation),\nthen\n[deploy in production](https://meltano.com/docs/production.html)\nalong with the built-in\n[Airflow integration](https://meltano.com/docs/production.html#airflow-orchestrator),\nor inside your\n[orchestrator of choice](https://meltano.com/docs/production.html#meltano-elt).\n\nMeltano [embraces](/docs/#embracing-singer) the [Singer](https://www.singer.io/) standard and its community-maintained library of open source\n[extractors](https://hub.meltano.com/extractors/) and\n[loaders](https://hub.meltano.com/loaders/),\nand leverages [dbt](https://www.getdbt.com) for [transformation](https://meltano.com/docs/transforms.html).\n\n## Documentation\n\nCheck out the ["Getting Started" guide](https://meltano.com/docs/getting-started.html)\nor find the full documentation at <https://www.meltano.com/docs/>.\n\n## Contributing to Meltano\n\nWe welcome contributions and improvements, please see the [contribution guidelines](https://meltano.com/docs/contributor-guide.html)\n\n## Responsible Disclosure Policy\n\nPlease refer to the [responsible disclosure policy](https://meltano.com/docs/responsible-disclosure.html) on our website.\n\n## License\n\nThis code is distributed under the MIT license, see the [LICENSE](LICENSE) file.\n',
    'author': 'Meltano Team and Contributors',
    'author_email': 'meltano@gitlab.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/meltano/meltano',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<3.10',
}


setup(**setup_kwargs)



"""Defines JobFinder."""

from datetime import datetime, timedelta

from .job import HEARTBEAT_VALID_MINUTES, HEARTBEATLESS_JOB_VALID_HOURS, Job, State


class JobFinder:
    """
    Query builder for the `Job` model for a certain `elt_uri`.
    """

    def __init__(self, job_id: str):
        self.job_id = job_id

    def latest(self, session):
        return (
            session.query(Job)
            .filter(Job.job_id == self.job_id)
            .order_by(Job.started_at.desc())
            .first()
        )

    def successful(self, session):
        return session.query(Job).filter(
            (Job.job_id == self.job_id)
            & (Job.state == State.SUCCESS)
            & Job.ended_at.isnot(None)
        )

    def running(self, session):
        """Find jobs in the running state."""
        return session.query(Job).filter(
            (Job.job_id == self.job_id) & (Job.state == State.RUNNING)
        )

    def latest_success(self, session):
        return self.successful(session).order_by(Job.ended_at.desc()).first()

    def latest_running(self, session):
        """Find the most recent job in the running state, if any."""
        return self.running(session).order_by(Job.started_at.desc()).first()

    def with_payload(self, session, flags=0, since=None):
        query = (
            session.query(Job)
            .filter(
                (Job.job_id == self.job_id)
                & (Job.payload_flags != 0)
                & (Job.payload_flags.op("&")(flags) == flags)
                & Job.ended_at.isnot(None)
            )
            .order_by(Job.ended_at.asc())
        )

        if since:
            query = query.filter(Job.ended_at > since)

        return query

    def latest_with_payload(self, session, **kwargs):
        return (
            self.with_payload(session, **kwargs)
            .order_by(None)  # Reset ascending order
            .order_by(Job.ended_at.desc())
            .first()
        )

    @classmethod
    def all_stale(cls, session):
        """Return all stale jobs."""
        now = datetime.utcnow()
        last_valid_heartbeat_at = now - timedelta(minutes=HEARTBEAT_VALID_MINUTES)
        last_valid_started_at = now - timedelta(hours=HEARTBEATLESS_JOB_VALID_HOURS)

        return session.query(Job).filter(
            (Job.state == State.RUNNING)
            & (
                (
                    Job.last_heartbeat_at.isnot(None)
                    & (Job.last_heartbeat_at < last_valid_heartbeat_at)
                )
                | (
                    Job.last_heartbeat_at.is_(None)
                    & (Job.started_at < last_valid_started_at)
                )
            )
        )

    def stale(self, session):
        """Return stale jobs with the instance's job ID."""
        return self.all_stale(session).filter(Job.job_id == self.job_id)



import asyncio
import logging
import os
import sys
from contextlib import contextmanager, redirect_stderr, redirect_stdout, suppress

import click
from async_generator import asynccontextmanager

from .utils import capture_subprocess_output, remove_ansi_escape_sequences


class OutputLogger:
    def __init__(self, file):
        self.file = file
        self.stdout = sys.stdout
        self.stderr = sys.stderr

        self.outs = {}
        self._max_name_length = None

    def out(self, name, stream=None, color=None):
        if stream in (None, sys.stderr):
            stream = self.stderr
        elif stream is sys.stdout:
            stream = self.stdout

        out = Out(self, name, file=self.file, stream=stream, color=color)
        self.outs[name] = out
        self._max_name_length = None

        return out

    @property
    def max_name_length(self):
        if self._max_name_length is None:
            name_lengths = [len(name) for name in self.outs.keys()]
            self._max_name_length = max(name_lengths, default=0)

        return self._max_name_length


class LineWriter:
    NEWLINE = "\n"

    def __init__(self, out):
        self.__out = out
        self.__skip_next_newline = False

    def __getattr__(self, name):
        return getattr(self.__out, name)

    def write(self, line):
        # Pre 3.7, `logging.StreamHandler.emit` will
        # write the message and terminator separately
        if self.__skip_next_newline:
            self.__skip_next_newline = False
            if line == self.NEWLINE:
                return

        if not line.endswith(self.NEWLINE):
            line = line + self.NEWLINE
            self.__skip_next_newline = True

        self.__out.writeline(line)


class FileDescriptorWriter:
    def __init__(self, out, fd):
        self.__out = out
        self.__writer = os.fdopen(fd, "w")

    def __getattr__(self, name):
        return getattr(self.__writer, name)

    def isatty(self):
        return self.__out.isatty()


class Out:
    """
    Simple Out class to log anything written in a stream to a file
     and then also write it to the stream.
    """

    def __init__(self, output_logger, name, file, stream, color="white"):
        self.output_logger = output_logger
        self.name = name
        self.color = color

        self.file = file
        self.stream = stream

        self.last_line = ""

        self._prefix = None
        self._max_name_length = None

    @property
    def prefix(self):
        max_name_length = self.output_logger.max_name_length
        if self._prefix is None or self._max_name_length != max_name_length:
            self._max_name_length = max_name_length

            padding = max(max_name_length, 6)
            padded_name = self.name.ljust(padding)
            self._prefix = click.style(f"{padded_name} | ", fg=self.color)

        return self._prefix

    @contextmanager
    def line_writer(self):
        yield LineWriter(self)

    @contextmanager
    def redirect_logging(self, format=None, ignore_errors=()):
        logger = logging.getLogger()
        original_log_handlers = logger.handlers

        line_writer = LineWriter(self)
        handler = logging.StreamHandler(line_writer)

        if not format:
            if logger.getEffectiveLevel() == logging.DEBUG:
                format = "%(levelname)s %(message)s"
            else:
                format = "%(message)s"

        formatter = logging.Formatter(fmt=format)
        handler.setFormatter(formatter)

        logger.handlers = [handler]

        try:
            yield
        except (KeyboardInterrupt, asyncio.CancelledError, *ignore_errors):
            raise
        except Exception as err:
            logger.error(str(err), exc_info=True)
            raise
        finally:
            logger.handlers = original_log_handlers

    @asynccontextmanager
    async def writer(self):
        read_fd, write_fd = os.pipe()

        reader = asyncio.ensure_future(self._read_from_fd(read_fd))
        writer = FileDescriptorWriter(self, write_fd)

        try:
            yield writer
        finally:
            writer.close()

            with suppress(asyncio.CancelledError):
                await reader

    @asynccontextmanager
    async def redirect_stdout(self):
        async with self.writer() as stdout:
            with redirect_stdout(stdout):
                yield

    @asynccontextmanager
    async def redirect_stderr(self):
        async with self.writer() as stderr:
            with redirect_stderr(stderr):
                yield

    def writeline(self, line):
        self.last_line = line

        click.echo(self.prefix + line, nl=False, file=self)
        self.flush()

    def write(self, data):
        self.stream.write(data)
        self.file.write(remove_ansi_escape_sequences(data))
        # always flush the file to keep it up to date with the stream
        self.file.flush()

    def flush(self):
        self.stream.flush()
        self.file.flush()

    def isatty(self):
        # Explicitly claim we're connected to a TTY to stop Click
        # from stripping ANSI codes
        return self.stream.isatty()

    async def _read_from_fd(self, read_fd):
        # Since we're redirecting our own stdout and stderr output,
        # the line length limit can be arbitrarily large.
        line_length_limit = 1024 * 1024 * 1024  # 1 GiB

        reader = asyncio.StreamReader(limit=line_length_limit)
        read_protocol = asyncio.StreamReaderProtocol(reader)

        loop = asyncio.get_event_loop()
        read_transport, _ = await loop.connect_read_pipe(
            lambda: read_protocol, os.fdopen(read_fd)
        )

        await capture_subprocess_output(reader, self)
